# -*- coding: utf-8 -*-
##############################################################################
# Name: peg_peg.py
# Purpose: PEG parser definition using PEG itself.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2009-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# PEG can be used to describe PEG.
# This example demonstrates building PEG parser using PEG based grammar of PEG
# grammar definition language.
##############################################################################

from __future__ import unicode_literals

import os
from arpeggio import *
from arpeggio.export import PMDOTExporter
from arpeggio.peg import PEGVisitor, ParserPEG


def main(debug=False):

    current_dir = os.path.dirname(__file__)
    peg_grammar = open(os.path.join(current_dir, 'peg.peg')).read()

    # ParserPEG will use ParserPython to parse peg_grammar definition and
    # create parser_model for parsing PEG based grammars
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    parser = ParserPEG(peg_grammar, 'peggrammar', debug=debug)

    # Now we will use created parser to parse the same peg_grammar used for
    # parser initialization. We can parse peg_grammar because it is specified
    # using PEG itself.
    print("PARSING")
    parse_tree = parser.parse(peg_grammar)

    # ASG should be the same as parser.parser_model because semantic
    # actions will create PEG parser (tree of ParsingExpressions).
    parser_model, comment_model = visit_parse_tree(
        parse_tree, PEGVisitor(root_rule_name='peggrammar',
                               comment_rule_name='comment',
                               ignore_case=False,
                               debug=debug))

    if debug:
        # This graph should be the same as peg_peg_parser_model.dot because
        # they define the same parser.
        PMDOTExporter().exportFile(parser_model,
                                   "peg_peg_new_parser_model.dot")

    # If we replace parser_mode with ASG constructed parser it will still
    # parse PEG grammars
    parser.parser_model = parser_model
    parser.parse(peg_grammar)

if __name__ == '__main__':
    main(debug=True)




# -*- coding: utf-8 -*-
#######################################################################
# Name: test_optional_in_choice
# Purpose: Optional matches always succeeds but should not stop alternative
#          probing on failed match.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2015 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals

# Grammar
from arpeggio import ParserPython, Optional, EOF

def g():    return [Optional('first'), Optional('second'), Optional('third')], EOF


def test_optional_in_choice():
    parser = ParserPython(g)
    input_str = "second"
    parse_tree = parser.parse(input_str)
    assert parse_tree is not None



#######################################################################
# Name: robot_peg.py
# Purpose: Simple DSL for defining robot movement (PEG version).
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2011-2017 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example is inspired by an example from LISA tool (http://labraj.uni-mb.si/lisa/)
# presented during the lecture given by prof. Marjan Mernik (http://lpm.uni-mb.si/mernik/)
# at the University of Novi Sad in June, 2011.
#
# An example of the robot program:
#    begin
#        up
#        up
#        left
#        down
#        right
#    end
#######################################################################
from __future__ import print_function, unicode_literals

import os
from arpeggio import visit_parse_tree
from arpeggio.cleanpeg import ParserPEG

# Semantic actions visitor
from robot import RobotVisitor


def main(debug=False):

    current_dir = os.path.dirname(__file__)

    # Load grammar
    robot_grammar = open(os.path.join(current_dir, 'robot.peg')).read()

    # First we will make a parser - an instance of the robot parser model.
    # Parser model is given in the form of PEG specification therefore we
    # are using ParserPEG class.
    parser = ParserPEG(robot_grammar, 'robot', debug=debug)

    # Load program code
    robot_program = open(os.path.join(current_dir, 'program.rbt')).read()

    # We create a parse tree out of textual input
    parse_tree = parser.parse(robot_program)

    # visit_parse_tree will start semantic analysis.
    # In this case semantic analysis will evaluate expression and
    # returned value will be the final position of the robot.
    return visit_parse_tree(parse_tree, RobotVisitor(debug=debug))


if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    print("position = ", main(debug=True))



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_semantic_action_results
# Purpose: Tests semantic action results passed to first_pass call
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ZeroOrMore, OneOrMore, ParserPython, \
    SemanticActionResults, PTNodeVisitor, visit_parse_tree
from arpeggio.export import PTDOTExporter
from arpeggio import RegExMatch as _


def grammar():      return first, "a", second
def first():        return [fourth, third], ZeroOrMore(third)
def second():       return OneOrMore(third), "b"
def third():        return [third_str, fourth]
def third_str():    return "3"
def fourth():       return _(r'\d+')


first_sar = None
third_sar = None


class Visitor(PTNodeVisitor):
    def visit_first(self, node, children):
        global first_sar
        first_sar = children

    def visit_third(self, node, children):
        global third_sar
        third_sar = children

        return 1


def test_semantic_action_results():

    global first_sar, third_sar

    input = "4 3 3 3 a 3 3 b"

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse(input)

    PTDOTExporter().exportFile(result, 'test_semantic_action_results_pt.dot')

    visit_parse_tree(result, Visitor())

    assert isinstance(first_sar, SemanticActionResults)
    assert len(first_sar.third) == 3
    assert third_sar.third_str[0] == '3'



from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ZeroOrMore, OneOrMore, ParserPython, EOF


def test_zeroormore_eolterm():

    def grammar():      return first, second, EOF
    def first():        return ZeroOrMore(["a", "b"], eolterm=True)
    def second():       return "a"

    # first rule should match only first line
    # so that second rule will match "a" on the new line
    input = """a a b a b b
    a"""

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse(input)

    assert result


def test_oneormore_eolterm():

    def grammar():      return first, second, EOF
    def first():        return OneOrMore(["a", "b"], eolterm=True)
    def second():       return "a"

    # first rule should match only first line
    # so that second rule will match "a" on the new line
    input = """a a a  b a
    a"""

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse(input)

    assert result



#######################################################################
# Name: simple.py
# Purpose: Simple language based on example from pyPEG
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2009-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example demonstrates grammar definition using python constructs.
# It is taken and adapted from pyPEG project (see http://www.fdik.org/pyPEG/).
#######################################################################

from __future__ import unicode_literals

import os
from arpeggio import *
from arpeggio import RegExMatch as _

# Grammar
def comment():          return [_("//.*"), _("/\*.*\*/")]
def literal():          return _(r'\d*\.\d*|\d+|".*?"')
def symbol():           return _(r"\w+")
def operator():         return _(r"\+|\-|\*|\/|\=\=")
def operation():        return symbol, operator, [literal, functioncall]
def expression():       return [literal, operation, functioncall]
def expressionlist():   return expression, ZeroOrMore(",", expression)
def returnstatement():  return Kwd("return"), expression
def ifstatement():      return Kwd("if"), "(", expression, ")", block, Kwd("else"), block
def statement():        return [ifstatement, returnstatement], ";"
def block():            return "{", OneOrMore(statement), "}"
def parameterlist():    return "(", symbol, ZeroOrMore(",", symbol), ")"
def functioncall():     return symbol, "(", expressionlist, ")"
def function():         return Kwd("function"), symbol, parameterlist, block
def simpleLanguage():   return function


def main(debug=False):

    # Load test program from file
    current_dir = os.path.dirname(__file__)
    test_program = open(os.path.join(current_dir, 'program.simple')).read()

    # Parser instantiation. simpleLanguage is the definition of the root rule
    # and comment is a grammar rule for comments.
    parser = ParserPython(simpleLanguage, comment, debug=debug)

    parse_tree = parser.parse(test_program)


if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




#######################################################################
# Name: calc.py
# Purpose: Simple expression evaluator example
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2009-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example demonstrates grammar definition using python constructs as
# well as using semantic actions to evaluate simple expression in infix
# notation.
#######################################################################

from __future__ import unicode_literals, print_function
try:
    text=unicode
except:
    text=str

from arpeggio import Optional, ZeroOrMore, OneOrMore, EOF, \
    ParserPython, PTNodeVisitor, visit_parse_tree
from arpeggio import RegExMatch as _

def number():     return _(r'\d*\.\d*|\d+')
def factor():     return Optional(["+","-"]), [number,
                          ("(", expression, ")")]
def term():       return factor, ZeroOrMore(["*","/"], factor)
def expression(): return term, ZeroOrMore(["+", "-"], term)
def calc():       return OneOrMore(expression), EOF


class CalcVisitor(PTNodeVisitor):

    def visit_number(self, node, children):
        """
        Converts node value to float.
        """
        if self.debug:
            print("Converting {}.".format(node.value))
        return float(node.value)

    def visit_factor(self, node, children):
        """
        Applies a sign to the expression or number.
        """
        if self.debug:
            print("Factor {}".format(children))
        if len(children) == 1:
            return children[0]
        sign = -1 if children[0] == '-' else 1
        return sign * children[-1]

    def visit_term(self, node, children):
        """
        Divides or multiplies factors.
        Factor nodes will be already evaluated.
        """
        if self.debug:
            print("Term {}".format(children))
        term = children[0]
        for i in range(2, len(children), 2):
            if children[i-1] == "*":
                term *= children[i]
            else:
                term /= children[i]
        if self.debug:
            print("Term = {}".format(term))
        return term

    def visit_expression(self, node, children):
        """
        Adds or subtracts terms.
        Term nodes will be already evaluated.
        """
        if self.debug:
            print("Expression {}".format(children))
        expr = children[0]
        for i in range(2, len(children), 2):
            if i and children[i - 1] == "-":
                expr -= children[i]
            else:
                expr += children[i]
        if self.debug:
            print("Expression = {}".format(expr))
        return expr


def main(debug=False):
    # First we will make a parser - an instance of the calc parser model.
    # Parser model is given in the form of python constructs therefore we
    # are using ParserPython class.
    parser = ParserPython(calc, debug=debug)

    # An expression we want to evaluate
    input_expr = "-(4-1)*5+(2+4.67)+5.89/(.2+7)"

    # We create a parse tree out of textual input_expr
    parse_tree = parser.parse(input_expr)

    # The result is obtained by semantic evaluation using visitor class.
    # visit_parse_tree will start semantic analysis.
    # In this case semantic analysis will evaluate expression and
    # returned value will be evaluated result of the input_expr expression.
    result = visit_parse_tree(parse_tree, CalcVisitor(debug=debug))

    # Check that result is valid
    assert (result - -7.51194444444) < 0.0001

    print("{} = {}".format(input_expr, result))

if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ZeroOrMore, OneOrMore, UnorderedGroup, \
    ParserPython, NoMatch, EOF


def test_zeroormore_with_separator():

    def grammar():
        return ZeroOrMore(['a', 'b'], sep=','), EOF

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse('a, b, b, b, a')
    assert result

    with pytest.raises(NoMatch):
        parser.parse('a, b a')


def test_oneormore_with_ordered_choice_separator():

    def grammar():
        return OneOrMore(['a', 'b'], sep=[',', ';']), EOF

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse('a, a; a,  b, a; a')
    assert result

    with pytest.raises(NoMatch):
        parser.parse('a, b a')

    with pytest.raises(NoMatch):
        parser.parse('a, b: a')


def test_unordered_group_with_separator():

    def grammar():
        return UnorderedGroup('a', 'b', 'c', sep=[',', ';']), EOF

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse('b , a, c')
    assert result
    result = parser.parse('b , c; a')
    assert result

    # Check separator matching
    with pytest.raises(NoMatch):
        parser.parse('a, b c')

    with pytest.raises(NoMatch):
        parser.parse('a, c: a')

    # Each element must be matched exactly once
    with pytest.raises(NoMatch):
        parser.parse('a, b, b; c')

    with pytest.raises(NoMatch):
        parser.parse('a, c')



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_python_parser
# Purpose: Test for parser constructed using Python-based grammars.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import Optional, ZeroOrMore, OneOrMore, EOF, ParserPython,\
    Sequence, NonTerminal
from arpeggio import RegExMatch as _


def number():     return _(r'\d*\.\d*|\d+')
def factor():     return Optional(["+", "-"]), [number,
                                                ("(", expression, ")")]
def term():       return factor, ZeroOrMore(["*", "/"], factor)
def expression(): return term, ZeroOrMore(["+", "-"], term)
def calc():       return OneOrMore(expression), EOF


def test_pp_construction():
    '''
    Tests parser construction from python internal DSL description.
    '''
    parser = ParserPython(calc)

    assert parser.parser_model.rule_name == 'calc'
    assert isinstance(parser.parser_model, Sequence)
    assert parser.parser_model.nodes[0].desc == 'OneOrMore'


def test_parse_input():

    parser = ParserPython(calc)
    input = "4+5*7/3.45*-45*(2.56+32)/-56*(2-1.34)"
    result = parser.parse(input)

    assert isinstance(result, NonTerminal)
    assert str(result) == "4 | + | 5 | * | 7 | / | 3.45 | * | - | 45 | * | ( | 2.56 | + | 32 | ) | / | - | 56 | * | ( | 2 | - | 1.34 | ) | "
    assert repr(result) == "[ [ [ [ number '4' [0] ] ],  '+' [1], [ [ number '5' [2] ],  '*' [3], [ number '7' [4] ],  '/' [5], [ number '3.45' [6] ],  '*' [10], [  '-' [11], number '45' [12] ],  '*' [14], [  '(' [15], [ [ [ number '2.56' [16] ] ],  '+' [20], [ [ number '32' [21] ] ] ],  ')' [23] ],  '/' [24], [  '-' [25], number '56' [26] ],  '*' [28], [  '(' [29], [ [ [ number '2' [30] ] ],  '-' [31], [ [ number '1.34' [32] ] ] ],  ')' [36] ] ] ], EOF [37] ]"




from arpeggio.cleanpeg import ParserPEG

def test_regex_with_empty_successful_match_in_repetition():
    grammar = \
            """
            rule = (subexpression)+
            subexpression = r'^.*$'
            """
    parser = ParserPEG(grammar, "rule")
    parsed = parser.parse("something simple")

    assert parsed.rule_name == "rule"
    assert parsed.subexpression.rule_name == "subexpression"



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_sequence_params
# Purpose: Test Sequence expression parameters.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, NoMatch, Sequence


def test_skipws():
    """
    skipws may be defined per Sequence.
    """

    def grammar():
        return Sequence("one", "two", "three"), "four"

    parser = ParserPython(grammar)

    # By default, skipws is True and whitespaces will be skipped.
    parser.parse("one two   three  four")

    def grammar():
        return Sequence("one", "two", "three", skipws=False), "four"

    parser = ParserPython(grammar)

    # If we disable skipws for sequence only then whitespace
    # skipping should not be done inside sequence.
    with pytest.raises(NoMatch):
        parser.parse("one two   three  four")

    # But it will be done outside of it
    parser.parse("onetwothree  four")


def test_ws():
    """
    ws can be changed per Sequence.
    """

    def grammar():
        return Sequence("one", "two", "three"), "four"

    parser = ParserPython(grammar)

    # By default, ws consists of space, tab and newline
    # So this should parse.
    parser.parse("""one
            two   three  four""")

    def grammar():
        return Sequence("one", "two", "three", ws=' '), "four"

    parser = ParserPython(grammar)

    # If we change ws per sequence and set it to space only
    # given input will raise exception
    with pytest.raises(NoMatch):
        parser.parse("""one
                two   three  four""")

    # But ws will be default outside of sequence
    parser.parse("""one two  three
        four""")

    # Test for ws with more than one char.
    def grammar():
        return Sequence("one", "two", "three", ws=' \t'), "four"

    parser = ParserPython(grammar)

    # If we change ws per sequence and set it to spaces and tabs
    # given input will raise exception
    with pytest.raises(NoMatch):
        parser.parse("one two   \nthree \t four")

    # But ws will be default outside of sequence
    parser.parse("one two  three \n\t four")

    # Inside sequence a spaces and tabs will be skipped
    parser.parse("one \t two\t three \nfour")



#-*- coding: utf-8 -*-
#######################################################################
# Testing parsing speed. This is used for the purpose of testing
#   of performance gains/loses for various approaches.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2016 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import print_function, unicode_literals

import codecs
import time
from os.path import dirname, join, getsize
from arpeggio import ParserPython
from grammar import rhapsody

def timeit(parser, file_name, message):
    print(message, 'File:', file_name)
    file_name = join(dirname(__file__), 'test_inputs', file_name)
    file_size = getsize(file_name)
    print('File size: {:.2f}'.format(file_size/1000), 'KB')

    with codecs.open(file_name, "r", encoding="utf-8") as f:
        content = f.read()

    t_start = time.time()
    parser.parse(content)
    t_end = time.time()

    print('Elapsed time: {:.2f}'.format(t_end - t_start), 'sec')
    print('Speed = {:.2f}'.format(file_size/1000/(t_end - t_start)), 'KB/sec')

    if parser.memoization:
        print('Cache hits = ', parser.cache_hits)
        print('Cache misses = ', parser.cache_misses)
        print('Success ratio[%] = ',
              parser.cache_hits*100/(parser.cache_hits + parser.cache_misses))
    print()


def main():

    # Small file
    file_name_small = 'LightSwitch.rpy'
    # Large file
    file_name_large = 'LightSwitchDouble.rpy'

    # No memoization
    parser = ParserPython(rhapsody)
    print('\n*** No memoization\n')
    for i in range(3):
        timeit(parser, file_name_small,
               '{}. Small file, no memoization.'.format(i + 1))
        timeit(parser, file_name_large,
               '{}. Large file, no memoization.'.format(i + 1))

    # Memoization
    parser = ParserPython(rhapsody, memoization=True)
    print('\n*** Memoization\n')
    for i in range(3):
        timeit(parser, file_name_small,
               '{}. Small file, with memoization.'.format(i + 1))
        timeit(parser, file_name_large,
               '{}. Large file, with memoization.'.format(i + 1))

if __name__ == '__main__':
    main()



from __future__ import unicode_literals
import pytest
from arpeggio import SemanticAction, ParserPython

def test_direct_rule_call():
    '''
    Test regression where in direct rule call semantic action is
    erroneously attached to both caller and callee.
    '''

    def grammar():  return rule1, rule2
    def rule1():    return "a"
    def rule2():    return rule1

    call_count = [0]

    class DummySemAction(SemanticAction):
        def first_pass(self, parser, node, nodes):
            call_count[0] += 1
            return SemanticAction.first_pass(self, parser, node, nodes)

    # Sem action is attached to rule2 only but
    # this bug will attach it to rule1 also resulting in
    # wrong call count.
    rule2.sem = DummySemAction()

    parser = ParserPython(grammar)
    parse_tree = parser.parse("aa")
    parser.getASG()

    assert call_count[0] == 1, "Semantic action should be called once!"



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_default_semantic_action
# Purpose: Default semantic action is applied during semantic analysis
#           if no action is given for node type. Default action converts
#           terminals to strings, remove StrMatch terminals from sequences.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa
from arpeggio import ParserPython, SemanticAction, ParseTreeNode
from arpeggio import RegExMatch as _

try:
    # For python 2.x
    text=unicode
except:
    # For python 3.x
    text=str

def grammar():      return parentheses, 'strmatch'
def parentheses():  return '(', rulea, ')'
def rulea():        return ['+', '-'], number
def number():       return _(r'\d+')


p_removed = False
number_str = False
parse_tree_node = False


class ParenthesesSA(SemanticAction):
    def first_pass(self, parser, node, children):
        global p_removed, parse_tree_node
        p_removed = text(children[0]) != '('
        parse_tree_node = isinstance(children[0], ParseTreeNode)
        return children[0] if len(children) == 1 else children[1]


class RuleSA(SemanticAction):
    def first_pass(self, parser, node, children):
        global number_str
        number_str = type(children[1]) == text
        return children[1]


parentheses.sem = ParenthesesSA()
rulea.sem = RuleSA()


def test_default_action_enabled():

    parser = ParserPython(grammar)

    parser.parse('(-34) strmatch')

    parser.getASG(defaults=True)

    assert p_removed
    assert number_str
    assert not parse_tree_node


def test_default_action_disabled():

    parser = ParserPython(grammar)

    parser.parse('(-34) strmatch')

    parser.getASG(defaults=False)

    assert not p_removed
    assert not number_str
    assert parse_tree_node



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_python_parser
# Purpose: Testing the dot exporter.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest
import os
from arpeggio.export import PMDOTExporter, PTDOTExporter

# Grammar
from arpeggio import Optional, ZeroOrMore, OneOrMore, EOF, ParserPython
from arpeggio import RegExMatch as _


def number():     return _(r'\d*\.\d*|\d+')
def factor():     return Optional(["+","-"]), [number,
                                               ("(", expression, ")")]
def term():       return factor, ZeroOrMore(["*","/"], factor)
def expression(): return term, ZeroOrMore(["+", "-"], term)
def calc():       return OneOrMore(expression), EOF


@pytest.fixture
def parser():
    return ParserPython(calc)


def test_export_parser_model(parser):
    """
    Testing parser model export
    """

    PMDOTExporter().exportFile(parser.parser_model,
                               "test_exporter_parser_model.dot")

    assert os.path.exists("test_exporter_parser_model.dot")


def test_export_parse_tree(parser):
    """
    Testing parse tree export.
    """

    parse_tree = parser.parse("-(4-1)*5+(2+4.67)+5.89/(.2+7)")
    PTDOTExporter().exportFile(parse_tree,
                               "test_exporter_parse_tree.dot")

    assert os.path.exists("test_exporter_parse_tree.dot")



#-*- coding: utf-8 -*-
#######################################################################
# Name: bibtex.py
# Purpose: Parser for bibtex files
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2013-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import print_function, unicode_literals

import pprint
import os
import sys
from arpeggio import *
from arpeggio import RegExMatch as _


# Grammar
def bibfile():                  return ZeroOrMore([comment_entry, bibentry, comment]), EOF
def comment_entry():            return "@comment", "{", _(r'[^}]*'), "}"
def bibentry():                 return bibtype, "{", bibkey, ",", field, ZeroOrMore(",", field), "}"
def field():                    return fieldname, "=", fieldvalue
def fieldvalue():               return [fieldvalue_braces, fieldvalue_quotes]
def fieldvalue_braces():        return "{", fieldvalue_braced_content, "}"
def fieldvalue_quotes():        return '"', fieldvalue_quoted_content, '"'

# Lexical rules
def fieldname():                return _(r'[-\w]+')
def comment():                  return _(r'[^@]+')
def bibtype():                  return _(r'@\w+')
def bibkey():                   return _(r'[^\s,]+')
def fieldvalue_quoted_content():    return _(r'((\\")|[^"])*')
def fieldvalue_braced_content():    return Combine(
                                            ZeroOrMore([Optional(And("{"), fieldvalue_inner),\
                                                        fieldvalue_part]))

def fieldvalue_part():          return _(r'((\\")|[^{}])+')
def fieldvalue_inner():         return "{", fieldvalue_braced_content, "}"


# Semantic actions visitor
class BibtexVisitor(PTNodeVisitor):

    def visit_bibfile(self, node, children):
        """
        Just returns list of child nodes (bibentries).
        """
        if self.debug:
            print("Processing Bibfile")

        # Return only dict nodes
        return [x for x in children if type(x) is dict]

    def visit_bibentry(self, node, children):
        """
        Constructs a map where key is bibentry field name.
        Key is returned under 'bibkey' key. Type is returned under 'bibtype'.
        """
        if self.debug:
            print("  Processing bibentry %s" % children[1])
        bib_entry_map = {
            'bibtype': children[0],
            'bibkey': children[1]
        }
        for field in children[2:]:
            bib_entry_map[field[0]] = field[1]
        return bib_entry_map

    def visit_field(self, node, children):
        """
        Constructs a tuple (fieldname, fieldvalue).
        """
        if self.debug:
            print("    Processing field %s" % children[0])
        field = (children[0], children[1])
        return field

    def visit_fieldvalue(self, node, children):
        """
        This example is used in practice at the University of Novi Sad.
        Thus, handle accented chars found in writtings in Serbian language.
        Remove braces. Remove newlines.
        """
        value = children[0]
        value = value.replace(r"\'{c}", "ć")\
                     .replace(r"\'{C}", "Ć")\
                     .replace(r"\v{c}", "č")\
                     .replace(r"\v{C}", "Č")\
                     .replace(r"\v{z}", "ž")\
                     .replace(r"\v{Z}", "Ž")\
                     .replace(r"\v{s}", "š")\
                     .replace(r"\v{S}", "Š")
        value = re.sub("[\n{}]", '', value)
        return value


def main(debug=False, file_name=None):
    # First we will make a parser - an instance of the bib parser model.
    # Parser model is given in the form of python constructs therefore we
    # are using ParserPython class.
    parser = ParserPython(bibfile, debug=debug)

    if not file_name:
        file_name = os.path.join(os.path.dirname(__file__),
                                 'bibtex_example.bib')

    with codecs.open(file_name, "r", encoding="utf-8") as bibtexfile:
        bibtexfile_content = bibtexfile.read()

    # We create a parse tree or abstract syntax tree out of
    # textual input
    parse_tree = parser.parse(bibtexfile_content)

    # visit_parse_tree will start semantic analysis.
    # In this case semantic analysis will return list of bibentry maps.
    ast = visit_parse_tree(parse_tree, BibtexVisitor(debug=debug))

    return ast

if __name__ == "__main__":
    # First parameter is bibtex file
    if len(sys.argv) > 1:
        # In debug mode dot (graphviz) files for parser model
        # and parse tree will be created for visualization.
        # Checkout current folder for .dot files.
        entries = main(debug=True, file_name=sys.argv[1])
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(entries)
    else:
        print("Usage: python bibtex.py file_to_parse")



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_error_reporting
# Purpose: Test error reporting for various cases.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2015 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import unicode_literals
import pytest

from arpeggio import Optional, Not, ParserPython, NoMatch, EOF
from arpeggio import RegExMatch as _


def test_non_optional_precedence():
    """
    Test that all tried match at position are reported.
    """
    def grammar():
        return Optional('a'), 'b'

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('c')
    assert "Expected 'a' or 'b'" in str(e.value)
    assert (e.value.line, e.value.col) == (1, 1)

    def grammar():
        return ['b', Optional('a')]

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('c')

    assert "Expected 'b'" in str(e.value)
    assert (e.value.line, e.value.col) == (1, 1)


def test_optional_with_better_match():
    """
    Test that optional match that has gone further in the input stream
    has precedence over non-optional.
    """

    def grammar():  return [first, Optional(second)]
    def first():    return 'one', 'two', 'three', '4'
    def second():   return 'one', 'two', 'three', 'four', 'five'

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('one two three four 5')

    assert "Expected 'five'" in str(e.value)
    assert (e.value.line, e.value.col) == (1, 20)


def test_alternative_added():
    """
    Test that matches from alternative branches at the same positiona are
    reported.
    """

    def grammar():
        return ['one', 'two'], _(r'\w+')

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('   three ident')
    assert "Expected 'one' or 'two'" in str(e.value)
    assert (e.value.line, e.value.col) == (1, 4)


def test_file_name_reporting():
    """
    Test that if parser has file name set it will be reported.
    """

    def grammar():      return Optional('a'), 'b', EOF

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse("\n\n   a c", file_name="test_file.peg")
    assert "Expected 'b' at position test_file.peg:(3, 6)" in str(e.value)
    assert (e.value.line, e.value.col) == (3, 6)


def test_comment_matching_not_reported():
    """
    Test that matching of comments is not reported.
    """

    def grammar():      return Optional('a'), 'b', EOF
    def comments():     return _(r'//.*$')

    parser = ParserPython(grammar, comments)

    with pytest.raises(NoMatch) as e:
        parser.parse('\n\n a // This is a comment \n c')
    assert "Expected 'b' at position (4, 2)" in str(e.value)
    assert (e.value.line, e.value.col) == (4, 2)


def test_not_match_at_beginning():
    """
    Test that matching of Not ParsingExpression is not reported in the
    error message.
    """

    def grammar():
        return Not('one'), _(r'\w+')

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('   one ident')
    assert "Not expected input" in str(e.value)


def test_not_match_as_alternative():
    """
    Test that Not is not reported if a part of OrderedChoice.
    """

    def grammar():
        return ['one', Not('two')], _(r'\w+')

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('   three ident')
    assert "Expected 'one' at " in str(e.value)


def test_sequence_of_nots():
    """
    Test that sequence of Not rules is handled properly.
    """

    def grammar():
        return Not('one'), Not('two'), _(r'\w+')

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('   two ident')
    assert "Not expected input" in str(e.value)


def test_compound_not_match():
    """
    Test a more complex Not match error reporting.
    """
    def grammar():
        return [Not(['two', 'three']), 'one', 'two'], _(r'\w+')

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch) as e:
        parser.parse('   three ident')
    assert "Expected 'one' or 'two' at" in str(e.value)

    with pytest.raises(NoMatch) as e:
        parser.parse('   four ident')
    assert "Expected 'one' or 'two' at" in str(e.value)



from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, ZeroOrMore, Sequence, OrderedChoice, \
    EOF, NoMatch


def test_ordered_choice_skipws_ws():

    # Both rules will skip white-spaces
    def sentence():
        return Sequence(ZeroOrMore(word), skipws=True), EOF

    def word():
        return OrderedChoice([(id, ' ', '.'), id, '.'], skipws=True)

    def id():
        return 'id'

    parser = ParserPython(sentence)

    # Thus this parses without problem
    # But the length is always 3 + EOF == 4
    # First alternative of word rule never matches
    tree = parser.parse("id id .")
    assert len(tree) == 4
    tree = parser.parse("id id.")
    assert len(tree) == 4
    tree = parser.parse("idid.")
    assert len(tree) == 4
    tree = parser.parse("idid .")
    assert len(tree) == 4

    # Now we change skipws flag
    def word():  # noqa
        return OrderedChoice([(id, ' ', '.'), id, '.'], skipws=False)

    parser = ParserPython(sentence)

    with pytest.raises(NoMatch):
        # This can't parse anymore
        parser.parse("id id .")

    tree = parser.parse("idid.")
    assert len(tree) == 4
    # This is the case where 'id .' will be matched by the first alternative of
    # word as there is no ws skipping
    tree = parser.parse("idid .")
    assert len(tree) == 3



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_pathologic_models
# Purpose: Test for grammar models that could lead to infinite loops are
#   handled properly.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import unicode_literals
import pytest

from arpeggio import ZeroOrMore, Optional, ParserPython, NoMatch, EOF


def test_optional_inside_zeroormore():
    """
    Test optional match inside a zero or more.
    Optional should always succeed thus inducing ZeroOrMore
    to try the match again.
    Arpeggio handle this case.
    """
    def grammar():
        return ZeroOrMore(Optional('a')), EOF

    parser = ParserPython(grammar)

    with pytest.raises(NoMatch):
        # This could lead to infinite loop
        parser.parse('b')



##############################################################################
# Name: json.py
# Purpose: Implementation of a simple JSON parser in arpeggio.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2009-2014 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example is based on jsonParser.py from the pyparsing project
# (see http://pyparsing.wikispaces.com/).
##############################################################################

from __future__ import unicode_literals

import os
from arpeggio import *
from arpeggio import RegExMatch as _

def TRUE():     return "true"
def FALSE():    return "false"
def NULL():     return "null"
def jsonString():       return '"', _('[^"]*'),'"'
def jsonNumber():       return _('-?\d+((\.\d*)?((e|E)(\+|-)?\d+)?)?')
def jsonValue():        return [jsonString, jsonNumber, jsonObject, jsonArray, TRUE, FALSE, NULL]
def jsonArray():        return "[", Optional(jsonElements), "]"
def jsonElements():     return jsonValue, ZeroOrMore(",", jsonValue)
def memberDef():        return jsonString, ":", jsonValue
def jsonMembers():      return memberDef, ZeroOrMore(",", memberDef)
def jsonObject():       return "{", Optional(jsonMembers), "}"
def jsonFile():         return jsonObject, EOF


def main(debug=False):
    # Creating parser from parser model.
    parser = ParserPython(jsonFile, debug=debug)

    # Load test JSON file
    current_dir = os.path.dirname(__file__)
    testdata = open(os.path.join(current_dir, 'test.json')).read()

    # Parse json string
    parse_tree = parser.parse(testdata)

    # parse_tree can now be analysed and transformed to some other form
    # using e.g. visitor support. See http://textx.github.io/Arpeggio/semantics/

if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




#-*- coding: utf-8 -*-
#######################################################################
# Name: grammar.py
# Purpose: Grammar for Rational Rhapsody. For testing purposes.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2016 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import print_function, unicode_literals

from arpeggio import *
from arpeggio import RegExMatch as _


# Grammar
def rhapsody():                 return header, obj
def header():                   return _(r'[^\n]*')
def obj():                      return '{', ident, OneOrMore(prop), '}'
def prop():                     return '-', ident, '=', \
                                       Optional(value, \
                                                ZeroOrMore(Optional(';'), Not(['-', '}']), value)), \
                                       Optional(';')
def value():                    return [_string, _int, _float, GUID, obj, ident]
def GUID():                     return 'GUID', _(r'[a-f0-9]*-[a-f0-9]*-[a-f0-9]*-[a-f0-9]*-[a-f0-9]*')

def ident():                    return _(r'[^\d\W]\w*\b')
def _string():                  return _(r'("(\\"|[^"])*")|(\'(\\\'|[^\'])*\')')
def _int():                     return _(r'[-+]?[0-9]+\b')
def _float():                   return _(r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\b')



# coding=utf-8
from __future__ import unicode_literals, absolute_import
from arpeggio import ParserPython, EOF
import pytest


def test_parser_resilience():
    """Tests that arpeggio parsers recover successfully from failure."""
    parser = ParserPython(('findme', EOF))
    with pytest.raises(TypeError):
        parser.parse(map)
    assert parser.parse('  findme  ') is not None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals, print_function

import pytest
from arpeggio.cleanpeg import ParserPEG

input = """\
<?php
class Enum {
    protected $self = array();
    public function __construct( $fun ) {
        $args = func_get_args();
        for( $i=0, $n=count($args); $i<$n; $i++ )
            $this->add($args[$i]);
    }
   
    public function __get(  $name = null ) {
        return $this->self[$name];
    }
   
    public function add(  $name = null,  $enum = null ) {
        if( isset($enum) )
            $this->self[$name] = $enum;
        else
            $this->self[$name] = end($this->self) + 1;
    }


"""

grammar = """

calc = test

test = visibility ws* function_keyword ws* word ws* arguments* ws*
function = visibility "function" word arguments block
block = "{" ws* r'[^}]*' ws* "}"
arguments = "(" ws* argument* ws* ")"

// $types = array("cappuccino")
// arguments end with optional comma
argument = ( byvalue / byreference ) ("=" value )* ","*
byreference = "&" byvalue
byvalue = variable

// value may be variable or array or string or any php type
value = variable

visibility = "public" / "protected" / "private"
function_keyword = "function"

variable = "$" literal r'[a-zA-Z0-9_]*'
word = r'[a-zA-Z0-9_]+'
literal = r'[a-zA-Z]+'

comment = r'("//.*")|("/\*.*\*/")'
symbol = r'[\W]+'

anyword = r'[\w]*' ws*
ws = r'[\s]+'


"""


def argument(parser, node, children):
    """
    Removes parenthesis if exists and returns what was contained inside.
    """
    print(children)

    if len(children) == 1:
        print(children[0])
        return children[0]

    sign = -1 if children[0] == '-' else 1

    return sign * children[-1]

# Rules are mapped to semantic actions
sem_actions = {
    "argument": argument,
}


def test_issue_16():

    parser = ParserPEG(grammar, "calc", skipws=False)

    input_expr = """public function __construct( )"""
    parse_tree = parser.parse(input_expr)

    # Do semantic analysis. Do not use default actions.
    asg = parser.getASG(sem_actions=sem_actions, defaults=False)

    assert asg



#-*- coding: utf-8 -*-
#######################################################################
# Purpose: Testing memory consumption with memoization enabled.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2016 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import print_function, unicode_literals

import codecs
from os.path import dirname, join
from memory_profiler import profile
from arpeggio import ParserPython
from grammar import rhapsody

@profile
def memoization():

    parser = ParserPython(rhapsody, memoization=True)

    # Smaller file
    file_name = join(dirname(__file__), 'test_inputs', 'LightSwitch.rpy')
    with codecs.open(file_name, "r", encoding="utf-8") as f:
        content = f.read()

    small = parser.parse(content)

    # File that is double in size
    file_name = join(dirname(__file__), 'test_inputs', 'LightSwitchDouble.rpy')
    with codecs.open(file_name, "r", encoding="utf-8") as f:
        content = f.read()

    large = parser.parse(content)

if __name__ == '__main__':
    memoization()



from __future__ import unicode_literals
import pytest
import sys
from arpeggio import ParserPython


def test_memoization_positive(capsys):
    '''
    Test that already matched rule is found in the cache on
    subsequent matches.
    Args:
        capsys - pytest fixture for output capture
    '''

    def grammar():  return [(rule1, ruleb), (rule1, rulec)]
    def rule1():    return rulea, ruleb
    def rulea():    return "a"
    def ruleb():    return "b"
    def rulec():    return "c"

    parser = ParserPython(grammar, memoization=True, debug=True)

    # Parse input where a rule1 will match but ruleb will fail
    # Second sequence will try rule1 again on the same location
    # and result should be found in the cache.
    parse_tree = parser.parse("a   b   c")

    # Assert that cached result is used
    assert  "Cache hit" in capsys.readouterr()[0]
    assert parser.cache_hits == 1
    assert parser.cache_misses == 4

def test_memoization_nomatch(capsys):
    '''
    Test that already failed match is found in the cache on
    subsequent matches.
    '''

    def grammar():  return [(rule1, ruleb), [rule1, rulec]]
    def rule1():    return rulea, ruleb
    def rulea():    return "a"
    def ruleb():    return "b"
    def rulec():    return "c"

    parser = ParserPython(grammar, memoization=True, debug=True)
    parse_tree = parser.parse("c")

    assert  "Cache hit for [rule1=Sequence, 0] = '0'" in capsys.readouterr()[0]
    assert parser.cache_hits == 1
    assert parser.cache_misses == 4





# -*- coding: utf-8 -*-
#######################################################################
# Name: test_suppression
# Purpose: Test suppresion of parse tree nodes.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2016 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa
from arpeggio import ParserPython, Sequence, StrMatch, RegExMatch


def test_sequence_suppress():
    """
    """

    def grammar():
        return Sequence("one", "two", "three", suppress=True), "four"

    parser = ParserPython(grammar)

    result = parser.parse("one two three four")
    assert result[0] == "four"


def test_suppress_string_match():
    """
    Test that string matches with suppress=True do not produce parse tree
    nodes.
    """

    class SuppressStrMatch(StrMatch):
        suppress = True

    def grammar():
        return "one", "two", SuppressStrMatch("three"), "four"

    parser = ParserPython(grammar)

    result = parser.parse("one two three four")
    assert len(result) == 3
    assert result[1] == "two"
    assert result[2] == "four"


def test_register_syntax_classes_suppress():
    """
    Test suppressing by overriding special syntax forms (lists - OrderedChoice,
    tuples - Sequences and string - StrMatch).
    """

    class SuppressStrMatch(StrMatch):
        suppress = True

    def grammar():
        return "one", "two", RegExMatch(r'\d+'), "three"

    parser = ParserPython(grammar,
                          syntax_classes={'StrMatch': SuppressStrMatch})

    result = parser.parse("one two 42 three")

    # Only regex will end up in the tree
    assert len(result) == 1
    assert result[0] == "42"



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_ptnode_navigation_expressions
# Purpose: Test ParseTreeNode navigation expressions.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ParserPython, ZeroOrMore, ParseTreeNode, NonTerminal


def foo(): return "a", bar, "b", baz, bar2, ZeroOrMore(bar)
def bar(): return [bla, bum], baz, "c"
def bar2():return ZeroOrMore(bla)
def baz(): return "d"
def bla(): return "bla"
def bum(): return ["bum", "bam"]


def test_lookup_single():

    parser = ParserPython(foo, reduce_tree=False)

    result = parser.parse("a bum d c b d bla bum d c")

    # Uncomment following line to visualize the parse tree in graphviz
    # PTDOTExporter().exportFile(result, 'test_ptnode_navigation_expressions.dot')

    assert isinstance(result, ParseTreeNode)
    assert isinstance(result.bar, NonTerminal)
    # dot access
    assert result.bar.rule_name == 'bar'
    # Index access
    assert result[1].rule_name == 'bar'

    # There are six children from result
    assert len(result) == 6

    # There is two bar matched from result (at the begging and from ZeroOrMore)
    # Dot access collect all NTs from the given path
    assert len(result.bar) == 2
    # Verify position
    assert result.bar[0].position == 2
    assert result.bar[1].position == 18

    # Multilevel dot access returns all elements from all previous ones.
    # For example this returns all bum from all bar in result
    assert len(result.bar.bum) == 2
    # Verify that proper bum are returned
    assert result.bar.bum[0].rule_name == 'bum'
    assert result.bar.bum[1].position == 18

    # Access to terminal
    assert result.bar.bum[-1][0].value == 'bum'
    assert result.bar2.bla[0].value == 'bla'

    # The same for all bla from all bar2
    assert len(result.bar2.bla) == 1

    assert hasattr(result, "bar")
    assert hasattr(result, "baz")

    # Test that accessing an invalid rule name raises AttributeError
    with pytest.raises(AttributeError):
        result.unexisting
